@@ -1,7 +1,6 @@
 from pyteomics import mzml
 import sys
 import numpy as np
-from scipy.interpolate import UnivariateSpline
 import matplotlib.pyplot as plt
 from multiprocessing import Manager,Pool
 from threading import Thread
@@ -14,8 +13,13 @@ import math
 import random as rd
 import IPython.display
 from copy import deepcopy
-import sklearn.metrics as met
 import pandas as pd
+from scipy.spatial.distance import euclidean
+from fastdtw import fastdtw
+from scipy.interpolate import interp1d
+import pickle as pkl
+import os
+import sklearn.metrics as met
 
 
 #Utility functions:
@@ -85,6 +89,7 @@ def take_closest(myList, myNumber):
         return before
 
 
+
 def printProgressBar(iteration, total, prefix='', suffix='', decimals=1, length=50, fill='█', printEnd="\r"):
     """
     Call in a loop to create terminal progress bar
@@ -121,101 +126,104 @@ def updateProgress(q, total,message = ""):
 
 def getIndexOfClosestValue(l,v):
     """
-    get index of list with closest value to v
+    get index of list with closest value to v, assumes l is sorted
     """
-    order = list(range(len(l)))
-    order.sort(key=lambda x:np.abs(l[x]-v))
-    return order[0]
-
-def generateSkylineFiles(peak_scores,peak_boundaries,samples,polarity,moleculeListName = "XCMS peaks"):
-    transitionList = deepcopy(peak_scores)
-
-    transitionList["Precursor Name"] = ["unknown " + str(index) for index, row in transitionList.iterrows()]
-    transitionList["Explicit Retention Time"] = [row["rt"] for index, row in
-                                                 transitionList.iterrows()]
-    polMapper = {"Positive": 1, "Negative": -1}
-    transitionList["Precursor Charge"] = [polMapper[polarity] for index, row in transitionList.iterrows()]
-    transitionList["Precursor m/z"] = [row["mz"] for index,row in transitionList.iterrows()]
-    transitionList["Molecule List Name"] = [moleculeListName for _ in range(len(transitionList))]
-    transitionList = transitionList[
-        ["Molecule List Name", "Precursor Name", "Precursor m/z", "Precursor Charge",
-         "Explicit Retention Time"]]
-
-    peakBoundariesSkyline = {}
-
-    # iterate through filenames and cpds
-    for fn in samples:
-        for index, row in transitionList.iterrows():
-            peakBoundariesSkyline[len(peakBoundariesSkyline)] = {"Min Start Time": peak_boundaries.at[index,fn][0],
-                                                   "Max End Time": peak_boundaries.at[index,fn][1],
-                                                   "Peptide Modified Sequence": row["Precursor Name"],
-                                                   "File Name": fn}
-
-    # format as df
-    peakBoundariesSkyline = pd.DataFrame.from_dict(peakBoundariesSkyline, orient="index")
-    peakBoundariesSkyline = peakBoundariesSkyline[["File Name", "Peptide Modified Sequence", "Min Start Time", "Max End Time"]]
-
-    return transitionList,peakBoundariesSkyline
-
-
-
+    pos = bisect_left(l, v)
+    if pos == 0:
+        return 0
+    if pos == len(l):
+        return pos - 1
+    before = l[pos - 1]
+    after = l[pos]
+    if after - v < v - before:
+        return pos
+    else:
+        return pos - 1
 
 
 class PeakDetective():
     """
-    Class for curation/detection of LC/MS peaks in untargerted metabolomics data
+    Class for curation/detection of LC/MS peaks in untargeted metabolomics data
     """
     def __init__(self,resolution=100,numCores=1,windowSize = 1.0):
         self.resolution = resolution
         self.numCores = numCores
         self.windowSize = windowSize
         self.smoother = Smoother(resolution)
-        self.classifier = Classifier(resolution)
+        self.classifier = ClassifierLatent(5)
         self.encoder = keras.Model(self.smoother.input, self.smoother.layers[7].output)
 
-    def plot_overlayedEIC(self,rawdatas,mz,rt_start,rt_end,smoothing=0,alpha=0.3):
+    def save(self,path):
+        if not os.path.exists(path): os.mkdir(path)
+        pkl.dump([self.resolution,self.windowSize],open(path+"parameters.pd","wb"))
+        self.smoother.save(path + "smoother.h5")
+        self.classifier.save(path+"classifier.h5")
+        self.encoder.save(path+"encoder.h5")
+
+
+    def load(self,path):
+        try:
+            [self.resouton,self.windowSize] = pkl.load(open(path + "parameters.pd","rb"))
+            self.smoother = keras.models.load_model(path+"smoother.h5")
+            self.classifier = keras.models.load_model(path+"classifier.h5")
+            self.encoder = keras.models.load_model(path+"encoder.h5")
+
+        except:
+            print("Error: pd and h5 files were not found, check path")
+
+
+
+    def plot_overlayedEIC(self,rawdatas,mz,rt_start,rt_end,alpha=0.3):
         ts = np.linspace(rt_start,rt_end,self.resolution)
         for data in rawdatas:
-            s = data.interpolate_data(mz,rt_start,rt_end,smoothing)
+            s = data.interpolate_data(mz,rt_start,rt_end)
             ints  = [np.max([x,0]) for x in s(ts)]
             plt.plot(ts,ints,alpha=alpha)
 
 
     @staticmethod
-    def getNormalizedIntensityVector(data,mzs,rtstarts,rtends,smoothing,resolution,q=None):
+    def getNormalizedIntensityVector(data,mzs,rtstarts,rtends,resolution,q=None):
         out = np.zeros((len(mzs),resolution))
         i=0
         for mz,rt_start,rt_end in zip(mzs,rtstarts,rtends):
-            s = data.interpolate_data(mz,rt_start,rt_end,smoothing)
+            s = data.interpolate_data(mz,rt_start,rt_end)
             out[i,:] = s(np.linspace(rt_start,rt_end,resolution))
             i += 1
         if type(q) != type(None):
             q.put(0)
         return out
 
-    def makeDataMatrix(self,rawdatas,mzs,rts,smoothing=0):
+    def makeDataMatrix(self,rawdatas,mzs,rts,align=False):
         rtstarts = [rt - self.windowSize/2 for rt in rts]
         rtends = [rt + self.windowSize/2 for rt in rts]
         args = []
         numToGetPerProcess = int(len(rawdatas)*len(mzs)/float(self.numCores))
+        featInds = []
         for rawdata in rawdatas:
             tmpMzs = []
             tmpRtStarts = []
             tmpRTends = []
-            for mz,rtstart,rtend in zip(mzs,rtstarts,rtends):
+            for mz,rtstart,rtend,i in zip(mzs,rtstarts,rtends,range(len(mzs))):
                 tmpMzs.append(mz)
                 tmpRtStarts.append(rtstart)
                 tmpRTends.append(rtend)
+                featInds.append(i)
                 if len(tmpMzs) == numToGetPerProcess:
-                    args.append([rawdata,tmpMzs,tmpRtStarts,tmpRTends,smoothing,self.resolution])
+                    args.append([rawdata,tmpMzs,tmpRtStarts,tmpRTends,self.resolution])
                     tmpMzs = []
                     tmpRtStarts = []
                     tmpRTends = []
 
-            if len(tmpMzs) > 0: args.append([rawdata, tmpMzs, tmpRtStarts, tmpRTends, smoothing,self.resolution])
+            if len(tmpMzs) > 0: args.append([rawdata, tmpMzs, tmpRtStarts, tmpRTends,self.resolution])
 
         result = startConcurrentTask(PeakDetective.getNormalizedIntensityVector, args, self.numCores, "forming matrix", len(args))
-        return np.concatenate(result,axis=0)
+
+        result = np.concatenate(result,axis=0)
+
+        if align:
+            result = alignDataMatrix(result,featInds,True,self.numCores)
+
+        return result
 
     def generateGaussianPeaks(self,n, centerDist, numPeaksDist=(0,2), widthFactor=0.1, heightFactor=1):
         X_signal = np.zeros((n,self.resolution)) #self.generateNoisePeaks(X_norm, tics)
@@ -236,7 +244,7 @@ class PeakDetective():
 
         return X_signal
 
-    def generateFalsePeaks(self,peaks,raw_datas, n=None):
+    def generateFalsePeaks(self,peaks,raw_datas, n=None,align=False):
         if type(n) == type(None):
             n = len(peaks)
 
@@ -244,20 +252,20 @@ class PeakDetective():
         peaks_rand["rt"] = rd.choices(list(peaks["rt"].values),k=n)
         peaks_rand["mz"] = rd.choices(list(peaks["mz"].values),k=n)
 
-        X_noise = self.makeDataMatrix(raw_datas,peaks_rand["mz"].values,peaks_rand["rt"].values)
+        X_noise = self.makeDataMatrix(raw_datas,peaks_rand["mz"].values,peaks_rand["rt"].values,align)
 
         return X_noise
 
-    def generateSignalPeaks(self,peaks,raw_datas,widthFactor = 0.1,heightFactor = 1,n=None):
+    def generateSignalPeaks(self,peaks,raw_datas,widthFactor = 0.1,heightFactor = 1,n=None,align=False):
         if type(n) == type(None):
             n = len(peaks)
-        X_noise = self.generateFalsePeaks(peaks,raw_datas,n=n)
+        X_noise = self.generateFalsePeaks(peaks,raw_datas,n=n,align=align)
         X_signal = self.generateGaussianPeaks(int(n*len(raw_datas)),[[0.45,0.5],[0.5,0.55]],(1,1),widthFactor,heightFactor)
 
         samp = peaks.loc[rd.choices(list(peaks.index.values),k=int(np.ceil(n))),:]
         mzs = list(samp["mz"].values)
 
-        tmp = self.makeDataMatrix(raw_datas,mzs,samp["rt"].values)
+        tmp = self.makeDataMatrix(raw_datas,mzs,samp["rt"].values,align=align)
         tmp = tmp[:n*len(raw_datas),:]
 
         signal_areas = np.array([integratePeak(x) for x in tmp])
@@ -291,7 +299,7 @@ class PeakDetective():
         self.encoder = keras.Model(smoother.input, smoother.layers[7].output)
         print("done")
 
-    def trainClassifier(self,X,y,X_val,y_val,class_epochs,batch_size,restarts):
+    def trainClassifier(self,X,y,X_val,y_val,min_epochs,max_epochs,batch_size,restarts):
 
         X_norm = normalizeMatrix(X)
         tics = np.log10(np.array([np.max([2, integratePeak(x)]) for x in X]))
@@ -301,6 +309,7 @@ class PeakDetective():
         tic_val = np.log10(np.array([np.max([2, integratePeak(x)]) for x in X_val]))
         X_val = self.encoder.predict(X_tmp)
 
+
         bestLoss = np.inf
         bestWeights = -1
         bestValErr = -1
@@ -324,13 +333,19 @@ class PeakDetective():
 
             classifer = ClassifierLatent(X_latent.shape[1])
 
-            classifer.fit([X_latent, tics], y, epochs=int(class_epochs),
+            if min_epochs > 0:
+                classifer.fit([X_latent,tics],y,epochs=int(min_epochs),batch_size=batch_size,verbose=0,
+                              validation_data=([X_val, tic_val], y_val))
+
+            classifer.fit([X_latent, tics], y, epochs=int(max_epochs-min_epochs),
                           batch_size=batch_size, verbose=0, callbacks=[cb, history],
                           validation_data=([X_val, tic_val], y_val))
 
             valLoss = history.history["val_loss"][cb.best_epoch]
             valErr = history.history["val_mean_absolute_error"][cb.best_epoch]
             bestEpoch = cb.best_epoch
+
+
             if valLoss < bestLoss:
                 bestLoss = valLoss
                 bestWeights = classifer.get_weights()
@@ -338,7 +353,7 @@ class PeakDetective():
                 bestBestEpoch = bestEpoch
                 trainLoss = history.history["loss"][cb.best_epoch]
                 trainErr = history.history["mean_absolute_error"][cb.best_epoch]
-        print("loss:",trainLoss,"mean_absolute_error:",trainErr,"val loss:", bestLoss, "val_mean_absolute_error:",bestValErr, "numEpochs:", bestBestEpoch)
+        print("loss:",trainLoss,"mean_absolute_error:",trainErr,"val loss:", bestLoss, "val_mean_absolute_error:",bestValErr, "numEpochs:", min_epochs + bestBestEpoch)
 
         classifer.set_weights(bestWeights)
 
@@ -346,7 +361,7 @@ class PeakDetective():
 
         return history
 
-    def trainClassifierActive(self,X,X_labeled,y_labeled,class_epochs,batch_size,restarts,numVal = 10,numManualPerRound=3,inJupyter=True):
+    def trainClassifierActive(self,X,X_labeled,y_labeled,min_epochs,max_epochs,batch_size,restarts,numVal = 10,numManualPerRound=3,inJupyter=True):
         trainingInds = []
 
         valInds = list(range(len(X)))
@@ -366,10 +381,11 @@ class PeakDetective():
             y_val[ind, 0] = 1 - val
             y_val[ind, 1] = val
 
+
         if len(X_labeled) > 0:
             self.trainClassifier(X_labeled,
                                  y_labeled,
-                                 X_val,y_val, class_epochs, batch_size, restarts)
+                                 X_val,y_val, min_epochs,max_epochs, batch_size, restarts)
 
         y[updatingInds] = self.classifyMatrix(X[updatingInds])
 
@@ -393,17 +409,23 @@ class PeakDetective():
                     trainingInds.append(ind)
                     updatingInds.remove(ind)
 
-                X_train = np.concatenate((X[trainingInds],X_labeled),axis=0)
-                y_train = np.concatenate((y[trainingInds],y_labeled),axis=0)
+                if len(X_labeled) > 0:
+                    X_train = np.concatenate((X[trainingInds],X_labeled),axis=0)
+                    y_train = np.concatenate((y[trainingInds],y_labeled),axis=0)
+                else:
+                    X_train = X[trainingInds]
+                    y_train = y[trainingInds]
 
                 if len(X_train) > 0:
-                    self.trainClassifier(X_train,y_train,X_val,y_val,class_epochs,batch_size,restarts)
+                    self.trainClassifier(X_train,y_train,X_val,y_val,min_epochs,max_epochs,batch_size,restarts)
 
                 y[updatingInds] = self.classifyMatrix(X[updatingInds])
 
                 plt.figure()
                 plt.hist(y[:, 1], bins=20)
                 plt.title("round" + str(i + 1))
+                plt.xlabel("PeakDetective Score")
+                plt.ylabel("Number of features")
                 plt.show()
 
                 print(str(len(updatingInds)) + " unclassified features remaining")
@@ -418,6 +440,8 @@ class PeakDetective():
                 doMore = False
             i += 1
 
+        plotScoringStatistics(self.classifyMatrix(X_val)[:,1],y_val[:,1])
+
 
     def classifyMatrix(self,X):
         # normalize matrix
@@ -428,13 +452,12 @@ class PeakDetective():
         y = self.classifier.predict([X_latent, peak_areas])
         return y
 
-    def curatePeaks(self,raw_datas,peaks,threshold=0.5):
+    def curatePeaks(self,raw_datas,peaks,threshold=0.5,align=False):
         print("generating EICs...")
         mzs = peaks["mz"].values
         rts =  peaks["rt"].values
 
-        X = self.makeDataMatrix(raw_datas,mzs,rts)
-        peak_areas = [integratePeak(x) for x in X]
+        X = self.makeDataMatrix(raw_datas,mzs,rts,align=align)
 
         y = self.classifyMatrix(X)
 
@@ -449,78 +472,138 @@ class PeakDetective():
             for index in peaks.index.values:
                 keys.append([raw.filename, index])
 
-        transitionLists = self.getPeakBoundaries(X, [raw.filename for raw in raw_datas], peak_scores, threshold)
-        for raw in raw_datas:
-            peak_intensities[raw.filename] = raw.integrateTargets(transitionLists[raw.filename])[raw.filename].values
-
-        for [file, index], score, intensity in zip(keys, y[:, 1], peak_areas):
+        for [file, index], score in zip(keys, y[:, 1]):
             peak_scores.at[index, file] = score
-            peak_intensities.at[index, file] = intensity
             val = 0
             if score > threshold:
                 val = 1
             peak_curated.at[index, file] = val
 
+        peak_intensities = self.performIntegration(X, [raw.filename for raw in raw_datas], peak_scores, threshold)
+
         return peak_curated,peak_scores,peak_intensities
 
-    def detectPeaks(self, rawDatas, cutoff=0.5, intensityCutoff = 100,numDataPoints=3,window=10, noiseCutoff=4):
+    def detectPeaks(self, rawDatas, cutoff=0.5, intensityCutoff = 100,numDataPoints=3,window=0.05,align=True,detectFrac=0.0):
         rois = self.roiDetection(rawDatas, intensityCutoff=intensityCutoff, numDataPoints=numDataPoints)
 
         print("generating all EICs from ROIs...")
-        peaks = []
         dt = self.windowSize / self.resolution
         tmpRes = int(math.ceil((rawDatas[0].rts[-1] - rawDatas[0].rts[0] + self.windowSize) / dt))
         oldRes = int(self.resolution)
         oldwindow = float(self.windowSize)
         self.resolution = tmpRes
         self.windowSize = rawDatas[0].rts[-1] + self.windowSize/2 - rawDatas[0].rts[0] + self.windowSize/2
-        rts = [rawDatas[0].rts[-1] + self.windowSize / 2 for _ in rois]
-        X_tot = self.makeDataMatrix(rawDatas, rois, rts, 0)
+        rts = [(rawDatas[0].rts[-1] + rawDatas[0].rts[0])/2 for _ in rois]
+        X_tot = self.makeDataMatrix(rawDatas, rois, rts,align)
         self.resolution = oldRes
         self.windowSize = oldwindow
 
-        numPoints = math.floor(float(tmpRes - self.resolution) / window)
 
-        X = np.zeros((int(numPoints * len(rois)), self.resolution))
+        numPoints = 0
+        rt = rawDatas[0].rts[0]
+        while(rt <= rawDatas[0].rts[-1]):
+            numPoints += 1
+            rt += window
+
+        X = np.zeros((int(numPoints * len(rois) * len(rawDatas)), self.resolution))
 
-        counter = 0
 
         mzs = []
         rts = []
         files = []
+        featInds = []
 
-        start = 0
-        end = self.resolution
+
+        counter = 0
+        rowCounter = 0
+        trueDt = tmpRes / (rawDatas[0].rts[-1] + self.windowSize/2 - rawDatas[0].rts[0] + self.windowSize/2)
+        stride = int(np.floor(window * trueDt))
 
         for rawData in rawDatas:
-            rt = float(rawData.rts[0])
+            i = 0
             for row in range(len(rois)):
+                rt = float(rawDatas[0].rts[0])
+                start = 0
+                end = start + self.resolution
                 for _ in range(numPoints):
-                    X[counter, :] = X_tot[row, start:end]
+                    if end >= X_tot.shape[1]:
+                        n = X_tot.shape[1] - start
+                        print(start,end,n,X.shape,X_tot.shape,numPoints)
+                        X[counter,:n]  = X_tot[rowCounter,start:]
+                    else:
+                        X[counter, :] = X_tot[rowCounter, start:end]
                     counter += 1
-                    start += window
-                    end += window
+                    start += stride
+                    end += stride
                     mzs.append(rois[row])
                     rts.append(rt)
-                    rt += dt * window
+                    rt += window
                     files.append(rawData.filename)
+                    featInds.append(i)
+                    i += 1
+                rowCounter += 1
 
-        tics = np.log10(np.array([np.max([2, integratePeak(x)]) for x in X]))
+        X = X[:counter]
 
-        y = self.classifyMatrix(X)[:, 1]
+        print(len(X),"EICs constructed for evaluation")
+
+        peak_areas = [integratePeak(x) for x in X]
+        toClassify = [x for x in range(len(peak_areas)) if peak_areas[x] > intensityCutoff]
+
+        peak_scores = pd.DataFrame(index=range(len(set(featInds))))
+
+        orderedMzs = []
+        orderedRts = []
+
+        for row in range(len(rois)):
+            rt = float(rawDatas[0].rts[0])
+            for _ in range(numPoints):
+                orderedMzs.append(rois[row])
+                orderedRts.append(rt)
+                rt += window
+
+        peak_scores["mz"] = orderedMzs
+        peak_scores["rt"] = orderedRts
+
+        for rawData in rawDatas:
+            peak_scores[rawData.filename] = 0.0
+
+        y = np.zeros(len(X))
+        y[toClassify] = self.classifyMatrix(X[toClassify])[:, 1]
         print("done")
-        for mz, rt,filename, score, tic in zip(mzs, rts,files, y, tics):
-            if score > cutoff and tic > noiseCutoff:
-                peaks.append([mz, rt,filename, score,tic])
 
-        if len(peaks) > 0:
-            peaks = pd.DataFrame(data=np.array(peaks), columns=["mz", "rt","sample","score","peak area"])
-        else:
-            peaks = pd.DataFrame(columns=["mz", "rt","sample","score","peak area"])
+        goodInds = []
+        for id,filename, score in zip(featInds,files, y):
+            if score > cutoff:
+                goodInds.append(id)
+            peak_scores.at[id,filename] = score
 
-        print(len(peaks), " peaks found")
+        detectNum = int(np.ceil(len(rawDatas) * detectFrac))
+        counts = {x:0 for x in list(set(goodInds))}
+        for x in goodInds:
+            counts[x] += 1
+        goodInds = [x for x in counts if counts[x] >= detectNum]
+        goodInds.sort()
 
-        return peaks, X
+
+        toKeep = []
+        for x in range(len(rawDatas)):
+            for g in goodInds:
+                toKeep.append(x*len(peak_scores) + g)
+
+        X = X[toKeep]
+
+        peak_scores = peak_scores.loc[goodInds,:]
+
+        peak_scores = peak_scores.reset_index()
+
+        print(len(peak_scores), " peaks found")
+
+        peak_intensities = self.performIntegration(X, [raw.filename for raw in rawDatas], peak_scores, cutoff)
+        #for raw in rawDatas:
+        #    peak_intensities[raw.filename] = raw.integrateTargets(transitionLists[raw.filename])[raw.filename].values
+
+        return peak_scores, peak_intensities, rois
 
     def label_peaks(self,raw_data,peaks,inJupyter = True):
         rt_starts = [row["rt"] - self.windowSize/2 for _,row in peaks.iterrows()]
@@ -536,14 +619,17 @@ class PeakDetective():
 
     def labelPeak(self,vecs,rt_start,rt_end,inJupyter,title=""):
         plt.figure()
+        plt.ion()
         xs = np.linspace(rt_start, rt_end, len(vecs[0]))
         [plt.plot(xs, vec) for vec in vecs]
         plt.xlabel("time (arbitrary)")
         plt.ylabel("intensity")
         plt.title(title)
-        plt.show()
+        plt.show(block=False)
+        plt.pause(0.001)
         print("Enter classification (1=True Peak, 0=Artifact): ")
         val = input()
+
         while not validateInput(val):
             print("invalid classification: ")
             val = input()
@@ -600,36 +686,50 @@ class PeakDetective():
         goodInds = [index for index,row in peakScores.iterrows() if float(len([x for x in samples if row[x] > cutoff])) / len(samples) > frac]
         return peakScores.loc[goodInds,:]
 
-    def getPeakBoundaries(self,X,samples,peakScores,cutoff,defaultWidth=0.5):
+    def performIntegration(self, X, samples, peakScores, cutoff, defaultWidth=0.5,smooth=False):
         i = 0
-        toFill = []
-        transitionLists = {}
-        for samp in samples:
-            peakBoundaries = pd.DataFrame(index=peakScores.index.values,columns=["mz","rt_start","rt_end"])
-            bounds = []
-            for index,row in peakScores.iterrows():
+        peak_areas = pd.DataFrame(index=peakScores.index.values,columns=["mz","rt"])
+        peak_areas["mz"] = peakScores["mz"].values
+        peak_areas["rt"] = peakScores["rt"].values
+        peak_areas[samples] = np.zeros(peakScores[samples].values.shape)
+        print("integrating peaks...")
+        for index,row in peakScores.iterrows():
+            inds = []
+            allInds = []
+            for n,samp in enumerate(samples):
+                ind = i + n * len(peakScores)
                 if row[samp] > cutoff:
-                    lb,rb = findPeakBoundaries(X[i])
-                    actualRts = np.linspace(row["rt"]-self.windowSize/2,row["rt"]+self.windowSize/2,self.resolution)
-                    bounds.append([actualRts[lb],actualRts[rb]])
-                else:
-                    bounds.append([-1,-1])
-                    toFill.append((index,samp))
-            peakBoundaries["mz"] = peakScores["mz"].values
-            peakBoundaries["rt"] = peakScores["rt"].values
-            peakBoundaries[["rt_start","rt_end"]] = deepcopy(np.array(bounds))
-            transitionLists[samp] = peakBoundaries
-
-        for index,samp in toFill:
-            widths = [transitionLists[x].at[index,"rt_end"] - transitionLists[x].at[index,"rt_start"] for x in transitionLists if transitionLists[x].at[index,"rt_start"] > 0 and transitionLists[x].at[index,"rt_end"] > 0]
-            centers = [np.mean([transitionLists[x].at[index,"rt_end"],transitionLists[x].at[index,"rt_start"]]) for x in transitionLists if transitionLists[x].at[index,"rt_start"] > 0 and transitionLists[x].at[index,"rt_end"] > 0]
-            if len(widths) > 0:
-                transitionLists[samp].at[index,"rt_start"] = np.mean(centers) - np.mean(widths)/2
-                transitionLists[samp].at[index,"rt_end"] = np.mean(centers) + np.mean(widths)/2
+                    inds.append(ind)
+                allInds.append(ind)
+            if len(inds) > 0:
+                tmp = X[inds].sum(axis=0)
+                if smooth: tmp = self.smoother.predict(normalizeMatrix(np.array([tmp])),verbose=0)[0]
+                lb,rb = findPeakBoundaries(tmp)
             else:
-                transitionLists[samp].at[index,"rt_start"] = transitionLists[samp].at[index,"rt"] - defaultWidth/2
-                transitionLists[samp].at[index,"rt_end"] = transitionLists[samp].at[index,"rt"] + defaultWidth/2
-        return transitionLists
+                lb = int(np.round(self.resolution / 2 - defaultWidth * self.resolution / 2))
+                rb = int(np.round(self.resolution/2 + defaultWidth * self.resolution/2))
+            for ind,sample in zip(allInds,samples):
+                peak_areas.at[index,sample] = integratePeak(X[ind],[lb,rb])
+            i += 1
+            printProgressBar(i, len(peakScores), "integrating peaks", printEnd="")
+
+        return peak_areas
+
+def plotScoringStatistics(scores,labels):
+    tpr = []
+    fdr = []
+    cutoffs = np.linspace(0,1.0,100)
+    for cutoff in cutoffs:
+        fdr.append(1-met.precision_score(labels,scores > cutoff,zero_division=0))
+        tpr.append(met.recall_score(labels,scores>cutoff,zero_division=0))
+    plt.scatter(cutoffs,tpr,c="black",label="TPR")
+    plt.plot(cutoffs,tpr,c="black")
+
+    plt.scatter(cutoffs,fdr,c="red",label="FDR")
+    plt.plot(cutoffs,fdr,c="red")
+    plt.legend()
+    plt.xlabel("cutoff")
+    plt.ylabel("performance")
 
 
 def validateInput(input):
@@ -712,27 +812,43 @@ class rawData():
         mz_start = mz - width
         mz_end = mz + width
         rts = [x for x in self.rts if x > rt_start and x < rt_end]
-        intensity = [np.sum([i for mz,i in self.data[rt].items() if mz > mz_start and mz < mz_end]) for rt in rts]
+        intensity = [np.max([0,np.sum([i for mz,i in self.data[rt].items() if mz > mz_start and mz < mz_end])]) for rt in rts]
         return rts,intensity
 
-    def integrateTargets(self,transitionList):
-        areas = []
-        for index,row in transitionList.iterrows():
-            rts,intensity = self.extractEIC(row["mz"],row["rt_start"],row["rt_end"])
-            area = np.trapz(intensity,rts)
-            areas.append(area)
-        transitionList[self.filename] = areas
-        return transitionList
+    # def integrateTargets(self,transitionList):
+    #     areas = []
+    #     for index,row in transitionList.iterrows():
+    #         rts,intensity = self.extractEIC(row["mz"],row["rt_start"],row["rt_end"])
+    #         area = np.trapz(intensity,rts)
+    #         areas.append(area)
+    #     transitionList[self.filename] = areas
+    #     return transitionList
 
-    def interpolate_data(self,mz,rt_start,rt_end,smoothing=1):
+    def interpolate_data(self,mz,rt_start,rt_end):
         rts,intensity = self.extractEIC(mz,rt_start,rt_end)
         if len(rts) > 3:
-            smoothing = smoothing * len(rts) * np.max(intensity)
-            s = UnivariateSpline(rts,intensity,ext=1,s=smoothing)
+            #smoothing = smoothing * len(rts) * np.max(intensity)
+            #s = UnivariateSpline(rts,intensity,ext=1,s=smoothing)
+            s = interp1d(rts,intensity,kind="linear",fill_value=0,bounds_error=False)
         else:
-            s = UnivariateSpline([0,5,10,15],[0,0,0,0],ext=1,s=smoothing)
+            #s = UnivariateSpline([0,5,10,15],[0,0,0,0],ext=1,s=smoothing)
+            s = interp1d([0,5,10,15],[0,0,0,0],kind="linear",fill_value=0,bounds_error=False)
+
         return s
 
+    def getMergedSpectrum(self,rtRange=None,intensityThresh=0,ppm=1):
+        if rtRange is None:
+            rtRange = [self.rts[0],self.rts[-1]]
+        spectra = [{mz:i for mz,i in self.data[rt].items() if i > intensityThresh} for rt in self.rts if rt > rtRange[0] and rt < rtRange[1]]
+        return mergeSpectra(spectra,ppm)
+
+    def correctMassShift(self,shift):
+        convertToDa = lambda x: (shift * x / 1e6)
+        self.data = {rt:{mz-convertToDa(mz):i for mz,i in spec.items()} for rt,spec in self.data.items()}
+
+
+
+
 def Smoother(resolution):
     # build autoencoder
     autoencoderInput = keras.Input(shape=(resolution,))
@@ -798,40 +914,6 @@ def Smoother(resolution):
 
     return autoencoder
 
-def Classifier(resolution):
-    descriminatorInput = keras.Input(shape=(resolution,))
-    ticInput = keras.Input(shape=(1,))
-
-    kernelsize = 3
-    stride = 2
-    max_norm_value = 2.0
-
-    x = layers.Reshape((resolution, 1))(descriminatorInput)
-    x = layers.Conv1D(2, kernelsize, strides=stride, activation='relu', kernel_constraint=max_norm(max_norm_value),
-                      kernel_initializer='he_uniform')(x)
-
-    x = layers.Conv1D(4, kernelsize, strides=stride, activation='relu', kernel_constraint=max_norm(max_norm_value),
-                      kernel_initializer='he_uniform')(x)
-
-    x = layers.Flatten()(x)
-
-    x = layers.Dense(20, activation="relu")(x)
-
-    x = keras.Model(descriminatorInput, x)
-
-    tic = keras.Model(ticInput, layers.Dense(1, activation="linear")(ticInput))
-
-    x = layers.concatenate([x.output, tic.output], axis=1)
-    x = layers.Dense(10, activation="relu")(x)
-    output = layers.Dense(2, activation="softmax")(x)
-
-    classifier = keras.Model([descriminatorInput, ticInput], output, name="discriminator")
-
-    classifier.compile(loss='binary_crossentropy', optimizer=keras.optimizers.Adam(1e-4),
-                          metrics=['mean_absolute_error'])
-
-    return classifier
-
 def ClassifierLatent(resolution):
     descriminatorInput = keras.Input(shape=(resolution,))
     ticInput = keras.Input(shape=(1,))
@@ -912,11 +994,146 @@ def findPeakBoundaries(peak):
 
     return lb,rb
 
-def integratePeak(peak):
-    lb,rb = findPeakBoundaries(peak)
+def integratePeak(peak,bounds=None):
+    if bounds is None:
+        lb,rb = findPeakBoundaries(peak)
+    else:
+        lb = bounds[0]
+        rb = bounds[1]
     if lb != rb:
-        area = np.trapz(peak[lb:rb],np.linspace(lb,rb,rb-lb))
+        try:
+            area = np.trapz(peak[lb:rb],np.linspace(lb,rb,rb-lb))
+        except:
+            print(lb,rb)
+            area = 0
     else:
         area = 0
     return area
 
+
+def alignPeaks(peaks,normalize=True,reference=0,q=None):
+    if normalize:
+        peaks_norm = [safeNormalize(x) for x in peaks]
+    else:
+        peaks_norm = deepcopy(peaks)
+
+    reference_peak = peaks_norm[reference]
+    ref = list(range(len(reference_peak)))
+
+    for x,peak in enumerate(peaks_norm):
+        if x > 0:
+            distance, path = fastdtw(reference_peak, peak, dist=euclidean)
+            xs = []
+            ys = []
+
+            prev = path[0][0]
+            tmp = [peaks[x][path[0][1]]]
+            for p1,p2 in path[1:]:
+                if p1 != prev:
+                    xs.append(prev)
+                    ys.append(np.mean(tmp))
+                    tmp = []
+                tmp.append(peaks[x][p2])
+                prev = p1
+            xs.append(prev)
+            ys.append(np.mean(tmp))
+
+
+            f = interp1d(xs,ys,fill_value=0.0,bounds_error=False)
+            peaks[x] = [f(i) for i in ref]
+
+
+    if type(q) != type(None):
+        q.put(0)
+
+    return peaks
+
+def alignDataMatrix(matrix,peakInds,normalize=True,numCores=1):
+    uniquePeaks = list(set(peakInds))
+    order = [[] for _ in uniquePeaks]
+    args = [[[],normalize,0] for _ in uniquePeaks]
+    for i,peak,ind in zip(range(len(matrix)),matrix,peakInds):
+        args[ind][0].append(peak)
+        order[ind].append(i)
+    result = startConcurrentTask(alignPeaks,args,numCores,"aligning EICs",len(uniquePeaks))
+    for peaks,inds in zip(result,order):
+        for peak,ind in zip(peaks,inds):
+            matrix[ind] = peak
+    return matrix
+
+
+def isoLock(refMasses,queryMasses,driftRange=0.005,toTest=50,targetDiff=0.0001):
+    bestDrift = 0
+    bestIsoPairs = 0
+    refMasses = list(refMasses)
+    refMasses.sort()
+
+    for drift in np.linspace(-1*driftRange,driftRange,toTest):
+        sep = 1.00336 + drift
+        massesToFind = refMasses + sep
+        numPairs = 0
+        for m in massesToFind:
+            best = take_closest(queryMasses,m)
+            if np.abs(best-m) < targetDiff:
+                numPairs += 1
+        if (numPairs == bestIsoPairs and drift < bestDrift) or numPairs > bestIsoPairs:
+            bestDrift = drift
+            bestIsoPairs = numPairs
+    print(bestIsoPairs,"isopairs found with drift",bestDrift)
+    return bestDrift
+
+def takeClosestInd(myList, myNumber):
+    """
+    Assumes myList is sorted. Returns closest value to myNumber.
+
+    If two numbers are equally close, return the smallest number.
+    """
+
+    pos = bisect_left(myList, myNumber)
+    if pos == 0:
+        return pos
+    if pos == len(myList):
+        return len(myList) - 1
+    before = myList[pos - 1]
+    after = myList[pos]
+    if after - myNumber < myNumber - before:
+        return pos
+    else:
+        return pos-1
+
+def mergeSpectra(spectra,ppm):
+    if len(spectra) > 0:
+        mergedSpectrumMzs = [x for x in spectra[0]]
+        mergedSpectrumMzs.sort()
+        mergedSpectrumInts = [spectra[0][x] for x in mergedSpectrumMzs]
+
+        if len(spectra) > 1:
+            fact = ppm / 1e6
+            for spectrum in spectra[1:]:
+                for mz,i in spectrum.items():
+
+                    if len(mergedSpectrumMzs) > 0:
+
+                        x = takeClosestInd(mergedSpectrumMzs,mz)
+                        delta = mz * fact
+
+                        if mergedSpectrumMzs[x] > mz - delta and mergedSpectrumMzs[x] < mz + delta:
+                            mergedSpectrumInts[x] += i
+                        else:
+                            if mz < mergedSpectrumMzs[x]:
+                                mergedSpectrumMzs.insert(x,mz)
+                                mergedSpectrumInts.insert(x,i)
+                            else:
+                                mergedSpectrumMzs.insert(x+1,mz)
+                                mergedSpectrumInts.insert(x+1,i)
+
+                    else:
+                        mergedSpectrumMzs.append(mz)
+                        mergedSpectrumInts.append(i)
+    else:
+        mergedSpectrumMzs = []
+        mergedSpectrumInts = []
+
+
+    return {mz:i for mz,i in zip(mergedSpectrumMzs,mergedSpectrumInts)}
+
